import functools
import os
import re
from collections import defaultdict
from dataclasses import dataclass
from datetime import date, datetime
from subprocess import STDOUT, CalledProcessError, check_output
from typing import (
	Any,
	Callable,
	DefaultDict,
	Dict,
	Iterator,
	List,
	NamedTuple,
	Optional,
	Sequence,
	Set,
	Tuple,
)

import markdown
import yaml
from colored import Fore, Style  # type: ignore
from markdown.extensions.wikilinks import WikiLinkExtension

colored_logs = True


class FsNode:
	def __init__(self, parent: Optional['DirNode'], dirPath: str, fileName: Optional[str]) -> None:
		self.parent = parent
		self.fileName: Optional[str] = fileName
		self.dirName: str = os.path.basename(dirPath)
		self.fullPath: str = (
			(os.curdir if dirPath == '' else dirPath)
			if isinstance(self, DirNode) or not isinstance(fileName, str)
			else os.path.join(dirPath, fileName)
		)
		self.shouldPublish: bool = False
		# These fields are populated later by engine:
		self.linksTo: List['FsNode'] = []
		self.env: dict[str, Any] = {}

	def __repr__(self) -> str:
		return self.colorize(self.fullPath)

	def colorize(self, r: str) -> str:
		if colored_logs:
			if self.shouldPublish:
				r = f'{Style.bold}{Fore.spring_green_2b}{r}{Style.reset}'
		return r

	def setNodeAsPublic(self) -> None:
		self.shouldPublish = True

	def makePublic(self) -> None:
		self.runOnFsNodeAndAscendantNodes(self, lambda fsNode: fsNode.setNodeAsPublic())

	def isParentGitRepo(self) -> bool:
		if self.parent is None:
			return DirNode.isPwdGitRepo()
		return self.parent.isInGitRepo

	@staticmethod
	def runOnFsNodeAndAscendantNodes(startingNode: 'FsNode', fn: Callable[['FsNode'], None]) -> None:
		def walk(node: FsNode) -> None:
			fn(node)
			if node.parent is not None:
				walk(node.parent)

		walk(startingNode)


class FileNode(FsNode):
	@staticmethod
	def construct(parent: Optional['DirNode'], dirPath: str, fileName: str) -> 'FileNode':
		"""Constructs an object of type FileNode or one of its subclasses.
		Check if fileName needs to be processed with pypage.
		    If is a Md page, we return a Md object.
		    If is a non-Md page, we return a NonMd object.
		If it is neither, we return a FileNode object.
		"""
		_, extension = FileNode.splitFileName(fileName)

		if extension == '.md':
			return Md(parent, dirPath, fileName)

		if '.py.' in fileName:
			pySubExtPos = fileName.find('.py.')
			remainingExt = fileName[pySubExtPos:]
			expectedRemainingExt = '.py' + extension
			if remainingExt == expectedRemainingExt:
				# The condition above passing indicates this is a NonMd page file.
				realName = fileName[:pySubExtPos]
				rectifiedFileName = realName + extension
				return NonMd(realName, rectifiedFileName, parent, dirPath, fileName)

		if extension == '.html':
			return PageNode(parent, dirPath, fileName)

		return FileNode(parent, dirPath, fileName)

	@staticmethod
	def splitFileName(fileName: str) -> Tuple[str, str]:
		return os.path.splitext(fileName)

	def __init__(self, parent: Optional['DirNode'], dirPath: str, fileName: str) -> None:
		"""Do not use this constructor directly. Use the static method construct instead."""
		super().__init__(parent, dirPath, fileName)
		baseName, extension = FileNode.splitFileName(fileName)
		self.absoluteFilePath: str = os.path.join(os.getcwd(), self.fullPath)
		self.fileName: str = fileName  # pyright: ignore
		self.extension: str = extension
		self.baseName: str = baseName
		self.realName: str = self.baseName  # to be overwritten selectively

	def isIndex(self) -> bool:
		# Index pages are `index.md` or `index[.py].html` files.
		return self.realName == 'index' and (self.extension in ('.md', '.html'))

	def getLinkName(self) -> str:
		if self.isIndex():
			return self.getParentDir().getRectifiedName()
		return self.realName

	@property
	def title(self) -> str:
		if 'title' in self.env:
			return self.env['title']
		return self.realName

	def isPyPage(self) -> bool:
		return isinstance(self, PyPageNode)

	def colorize(self, r: str) -> str:
		if colored_logs:
			if self.isPyPage() is not None and self.shouldPublish:
				r = f'{Fore.spring_green_1}{r}{Style.reset}'
			elif isinstance(self, Md):
				r = f'{Fore.purple_4b}{r}{Style.reset}'
		return r

	def __repr__(self) -> str:
		r = f'{self.fileName}'
		r = self.colorize(r)
		r = super().colorize(r)
		return r

	def getParentDir(self) -> 'DirNode':
		# Note: The parent of a FileNode is always a DirNode (and never None).
		assert isinstance(self.parent, DirNode)
		parentDir: DirNode = self.parent
		return parentDir


class DirNode(FsNode):
	def __init__(
		self,
		parent: Optional['DirNode'],
		dirPath: str,
		# shouldIgnore(name: str, parentPath: str, isDir: bool) -> bool
		shouldIgnore: Callable[[str, str, bool], bool],
	) -> None:
		_, subDirNames, fileNames = next(os.walk(dirPath))
		dirPath = '' if dirPath == os.curdir else dirPath
		super().__init__(parent, dirPath, None)

		self.files: List[FileNode] = [
			FileNode.construct(self, dirPath, fileName)
			for fileName in fileNames
			if not shouldIgnore(fileName, self.fullPath, False)
		]
		self.subDirs: List[DirNode] = [
			DirNode(self, os.path.join(dirPath, subDirName), shouldIgnore)
			for subDirName in subDirNames
			if not shouldIgnore(subDirName, self.fullPath, True)
		]

	def getRectifiedName(self) -> str:
		# Note: if `dirName` is an empty string (""), that means we're at the root (/).
		return self.dirName if len(self.dirName) > 0 else '/'

	@functools.cached_property
	def isInGitRepo(self) -> bool:
		# We recheck if it's a git repo for each directory, in case
		# there's a symlink to a directory outside a git repo.
		return DirNode.isPwdGitRepo()

	@staticmethod
	def isPwdGitRepo() -> bool:
		try:
			check_output(['git', 'status'], stderr=STDOUT).decode()
			return True
		except CalledProcessError:
			return False

	@property
	def pages(self) -> Sequence['PageNode']:
		return [f for f in self.files if (isinstance(f, PageNode) and not f.isIndex())]

	def getPyPagesOtherThanIndex(self) -> Iterator['PyPageNode']:
		return (f for f in self.files if (isinstance(f, PyPageNode) and not f.isIndex()))

	@functools.cached_property
	def indexPage(self) -> Optional['PageNode']:
		indexFilter = filter(lambda f: f.isIndex(), self.files)
		indexFile: Optional[FileNode] = next(indexFilter, None)
		if indexFile:
			assert isinstance(indexFile, PageNode)
			return indexFile
		return None

	@property
	def title(self) -> Optional[str]:
		if self.indexPage and 'title' in self.indexPage.env:
			return self.indexPage.title
		return None

	@staticmethod
	def _displayDir(dirNode: 'DirNode', indent: int = 0) -> str:
		return (
			(' ' * 2 * indent)
			+ f'{dirNode} -> {dirNode.files}\n'
			+ ''.join(DirNode._displayDir(subDir, indent + 1) for subDir in dirNode.subDirs)
		)

	def displayDir(self) -> str:
		return self._displayDir(self)


class AltezaException(Exception):
	"""Alteza Exceptions"""


class PageNode(FileNode):
	default_date_format: str = '%Y %b %-d'
	default_datetime_format: str = default_date_format + ' at %-H:%M %p'

	def getLastModified(self, f: str = default_datetime_format) -> str:
		# The formatting below might only work on Linux. https://stackoverflow.com/a/29980406/908430
		return self.lastModifiedObj.strftime(f)

	def getIdeaDate(self, f: str = default_date_format) -> str:
		ideaDate = self.getIdeaDateObj()
		return ideaDate.strftime(f) if ideaDate else ''

	def getIdeaDateObj(self) -> Optional[date]:
		if isinstance(self, Md):
			if self.ideaDate is not None:
				return self.ideaDate
		return self.gitFirstAuthDate

	def getCreateDate(self, f: str = default_date_format) -> str:
		createDate = self.gitFirstAuthDate
		return createDate.strftime(f) if createDate else ''

	def getCreateDateObj(self) -> Optional[date]:
		return self.gitFirstAuthDate

	@functools.cached_property
	def lastModifiedObj(self) -> datetime:
		"""Get last modified date from: (a) git history, or (b) system modified time."""
		path = self.fileName
		if self.isParentGitRepo():
			lastUpdated = PageNode.getGitFileLastAuthDate(path)
			if lastUpdated is not None:
				return lastUpdated
		return datetime.fromtimestamp(os.path.getmtime(path))

	@staticmethod
	def getGitFileLastAuthDate(path: str) -> Optional[datetime]:
		try:
			git_output = check_output(['git', 'log', '-n', '1', '--pretty=format:%aI', path]).decode()
			return datetime.fromisoformat(git_output)
		except Exception:
			return None

	@functools.cached_property
	def gitFirstAuthDate(self) -> Optional[date]:
		path = self.fileName
		if self.isParentGitRepo():
			try:
				git_output = check_output(['git', 'log', '--reverse', '--pretty=format:%aI', path]).decode()
				first_commit_date = git_output.splitlines()[0]
				return datetime.fromisoformat(first_commit_date).date()
			except Exception:
				return None
		return None

	def __getattr__(self, attr: str) -> None:
		"""Allows for checking whether page.some_property exists more easily (without `hasattr`)."""
		return None


class PyPageNode(PageNode):
	temporal_link: Optional[Callable[[str], str]] = None

	def __init__(self, parent: Optional[DirNode], dirPath: str, fileName: str) -> None:
		super().__init__(parent, dirPath, fileName)
		self._pyPageOutput: Optional[str] = None  # to be generated (by pypage)

	@property
	def output(self) -> str:
		if self._pyPageOutput is None:
			raise AltezaException('PyPage output has not been generated yet.')
		assert isinstance(self._pyPageOutput, str)
		return self._pyPageOutput

	@output.setter
	def output(self, htmlOutput: str) -> None:
		self._pyPageOutput = htmlOutput


def buildWikiUrl(label: str, base: str, end: str) -> str:
	# pylint: disable=unused-argument
	if PyPageNode.temporal_link is None:
		raise AltezaException('PyPageNode.temporal_link is not set.')
	# pylint: disable=not-callable
	return PyPageNode.temporal_link(label)


class Md(PyPageNode):
	md = markdown.Markdown(
		# See: https://python-markdown.github.io/extensions/
		extensions=[
			# Extra extensions:
			'abbr',
			'attr_list',
			'def_list',
			'fenced_code',
			'footnotes',
			'md_in_html',
			'tables',
			# Standard extensions:
			'admonition',
			'codehilite',
			'meta',
			'mdx_breakless_lists',
			# "sane_lists",
			'mdx_truly_sane_lists',
			'smarty',  # not sure
			'toc',
			WikiLinkExtension(html_class='', build_url=buildWikiUrl),
		],
		extension_configs={'mdx_truly_sane_lists': {'nested_indent': 4}},
	)

	def __init__(self, parent: Optional[DirNode], dirPath: str, fileName: str) -> None:
		super().__init__(parent, dirPath, fileName)

		self.ideaDate: Optional[date] = None
		# Handle file names that start with a date:
		dateFragmentLength = len('YYYY-MM-DD-')
		if len(self.baseName) > dateFragmentLength:
			dateFragment_ = self.baseName[:dateFragmentLength]
			remainingBasename = self.baseName[dateFragmentLength:]
			if re.match('[0-9]{4}-[0-9]{2}-[0-9]{2}-$', dateFragment_):
				dateFragment = dateFragment_[:-1]
				self.ideaDate = date.fromisoformat(dateFragment)
				self.realName: str = remainingBasename

	class Result(NamedTuple):
		metadata: Dict[str, str]
		html: str

	@staticmethod
	def processMarkdown(text: str) -> Result:
		html: str = Md.md.convert(text)
		yamlFrontMatter: str = ''

		for name, lines in Md.md.Meta.items():  # type: ignore # pylint: disable=no-member
			yamlFrontMatter += f'{name} : {lines[0]} \n'
			for line in lines[1:]:
				yamlFrontMatter += ' ' * (len(name) + 3) + line + '\n'

		metadata = yaml.safe_load(yamlFrontMatter)
		if metadata is None:
			metadata = {}
		if not isinstance(metadata, dict):
			raise AltezaException('Expected yaml.safe_load to return a dict or None.')

		return Md.Result(metadata=metadata, html=html)


class NonMd(PyPageNode):
	def __init__(
		# pylint: disable=too-many-arguments, too-many-positional-arguments
		self,
		realName: str,
		rectifiedFileName: str,
		parent: Optional[DirNode],
		dirPath: str,
		fileName: str,
	) -> None:
		super().__init__(parent, dirPath, fileName)
		self.realName = realName
		self.rectifiedFileName: str = rectifiedFileName


class NameRegistry:
	def __init__(self, root: DirNode, skipForRegistry: Callable[[str], bool]) -> None:
		allFilesMulti: DefaultDict[str, Set[FileNode]] = defaultdict(set)

		def walk(node: DirNode) -> None:
			for fileNode in node.files:
				if not skipForRegistry(fileNode.fileName):
					allFilesMulti[fileNode.getLinkName()].add(fileNode)
			for d in node.subDirs:
				walk(d)

		walk(root)

		self.allFiles: Dict[str, FileNode] = {}
		for name, fileNodes in allFilesMulti.items():
			assert len(fileNodes) >= 1
			if len(fileNodes) > 1:
				self.errorOut(name, fileNodes)
			self.allFiles[name] = fileNodes.pop()

	def lookup(self, name: str) -> FileNode:
		if name not in self.allFiles:
			print(f'Link error: `{name}` was not found in the name registry.')
			raise AltezaException(f'Link error: {name}')

		return self.allFiles[name]

	@staticmethod
	def errorOut(name: str, fileNodes: Set[FileNode]) -> None:
		raise AltezaException(
			f"Error: The name '{name}' has multiple matches:\n"
			+ '  \n'.join(f' {fileNode.fullPath}' for fileNode in fileNodes)
		)

	def __repr__(self) -> str:
		return (
			'Name Registry:\n  '
			+ '\n  '.join(f'{k}: {v} {Fore.grey_39}@ {v.fullPath}{Style.reset}' for k, v in self.allFiles.items())
			+ '\n'
		)


@dataclass
class FsCrawlResult:
	rootDir: DirNode
	nameRegistry: NameRegistry


class Fs:
	configFileName: str = '__config__.py'
	ignoreAbsPaths: List[str] = []

	@staticmethod
	def readfile(file_path: str) -> str:
		with open(file_path, 'r', encoding='utf-8') as someFile:
			return someFile.read()

	@staticmethod
	def isHidden(name: str) -> bool:
		return name.startswith('.')

	@staticmethod
	def shouldIgnoreStandard(name: str) -> bool:
		if Fs.isHidden(name):
			return True
		if name in {'__pycache__'}:
			return True
		_, fileExt = os.path.splitext(name)
		if fileExt == '.pyc':
			return True
		if name != Fs.configFileName and fileExt == '.py':
			return True
		return False

	@staticmethod
	def defaultShouldIgnore(name: str, parentPath: str, isDir: bool) -> bool:
		# pylint: disable=unused-argument
		if Fs.shouldIgnoreStandard(name):
			return True

		fullPath = os.path.abspath(os.path.join(parentPath, name))
		for ignoreAbsPath in Fs.ignoreAbsPaths:
			if ignoreAbsPath in fullPath:
				return True

		return False

	@staticmethod
	def defaultSkipForRegistry(name: str) -> bool:
		if name == Fs.configFileName:
			return True
		return False

	@staticmethod
	def crawl(
		# Signature -- shouldIgnore(name: str, parentPath: str, isDir: bool) -> bool
		shouldIgnore: Callable[[str, str, bool], bool] = defaultShouldIgnore,
		skipForRegistry: Callable[[str], bool] = defaultSkipForRegistry,
	) -> FsCrawlResult:
		"""
		Crawl the current directory. Construct & return an FsNode tree and NameRegistry.
		"""
		dirPath: str = os.curdir

		rootDir: DirNode = DirNode(None, dirPath, shouldIgnore)
		nameRegistry = NameRegistry(rootDir, skipForRegistry)

		return FsCrawlResult(rootDir, nameRegistry)



import contextlib
import itertools
import json
import logging
import os
import shutil
import signal
import sys
import time
import types
from typing import Optional, Generator, List, Union, Dict, Set, Any

import sh  # type: ignore
from pypage import pypage, PypageError  # type: ignore
from tap import Tap
from watchdog.events import FileSystemEventHandler, FileSystemEvent, DirModifiedEvent
from watchdog.observers import Observer as WatchdogObserver

from .fs import (
	FsNode,
	FileNode,
	DirNode,
	NameRegistry,
	AltezaException,
	Md,
	NonMd,
	FsCrawlResult,
	Fs,
	Fore,
	Style,
	PageNode,
	PyPageNode,
)
from .version import version as alteza_version


class Args(Tap):  # pyre-ignore[13]
	content: str  # Directory to read the input content from.
	output: str  # Directory to write the generated site to.
	clear_output_dir: bool = False  # Delete the output directory, if it already exists.
	copy_assets: bool = False  # Copy static assets instead of symlinking to them.
	seed: str = '{}'  # Seed JSON data to add to the initial root env.
	watch: bool = False  # Watch for content changes, and rebuild.
	ignore: List[str] = []  # Paths to completely ignore.


class Content:
	def __init__(self, args: Args, fs: FsCrawlResult) -> None:
		self.inTemplate: bool = False
		self.templateCache: Dict[str, str] = {}
		self.seenTemplateLinks: Set[FileNode] = set()
		self.rootDir: DirNode = fs.rootDir
		self.nameRegistry: NameRegistry = fs.nameRegistry
		self.seed: Dict[str, Any] = json.loads(args.seed)
		self.fixSysPath()

	def link(self, srcFile: FileNode, dstFile: FileNode, pathOnly: bool = False) -> str:
		if not pathOnly:
			srcFile.linksTo.append(dstFile)  # This is used to determine reachability.
			if dstFile not in self.seenTemplateLinks:
				print(
					' ' * (4 if self.inTemplate else 2)
					+ f'{Fore.grey_42}Linking to:{Style.reset} {dstFile.getLinkName()}',
				)
				if self.inTemplate:
					self.seenTemplateLinks.add(dstFile)

		dstFileName = self.getFileUrlName(dstFile)

		srcPath = self.splitPath(srcFile.fullPath)[:-1]
		dstPath = self.splitPath(dstFile.fullPath)[:-1]
		commonLevel = 0
		for i in range(min(len(srcPath), len(dstPath))):
			if srcPath[i] == dstPath[i]:
				commonLevel += 1
		remainingPath = dstPath[commonLevel:] + [dstFileName]

		relativePath: List[str] = []
		if commonLevel < len(srcPath):
			stepsDown = len(srcPath) - commonLevel
			for _ in range(stepsDown):
				relativePath.append('..')
		for p in remainingPath:
			relativePath.append(p)
		if isinstance(srcFile, Md) and not srcFile.isIndex() and not pathOnly:
			relativePath = ['..'] + relativePath

		relativePathStr = os.path.join('', *relativePath)
		return relativePathStr

	@staticmethod
	def getFileUrlName(dstFile: FileNode) -> str:
		if dstFile.isIndex():
			return ''
		if isinstance(dstFile, Md):
			return dstFile.realName
		if isinstance(dstFile, NonMd):
			return dstFile.rectifiedFileName
		return dstFile.fileName

	def linkFlex(
		self,
		fromPyPage: PyPageNode,
		destination: Union[str, FileNode],
		pathOnly: bool = False,
	) -> str:
		if isinstance(destination, str):
			dstFile: FileNode = self.nameRegistry.lookup(destination)
			return self.link(fromPyPage, dstFile, pathOnly)
		if isinstance(destination, FileNode):
			return self.link(fromPyPage, destination, pathOnly)
		if isinstance(destination, DirNode):
			if not destination.indexPage:
				raise AltezaException(f'Directory `{destination}` has no index page.')
			return self.link(fromPyPage, destination.indexPage, pathOnly)
		raise AltezaException(f'Unknown link destination type: `{type(destination)}`.')

	def invokePyPage(self, pyPageNode: PyPageNode, env: dict[str, Any]) -> None:
		print(f'{Fore.gold_1}Processing:{Style.reset}', pyPageNode.fullPath)
		env = env.copy()

		# Enrich with current file:
		env |= {'file': pyPageNode}

		rawPyPageFileText: str
		if isinstance(pyPageNode, (Md, NonMd)):
			rawPyPageFileText = Fs.readfile(pyPageNode.absoluteFilePath)
		else:
			raise AltezaException(f'{pyPageNode} Unsupported type of PyPageNode.')

		def link(destination: Union[str, FileNode]) -> str:
			return self.linkFlex(pyPageNode, destination)

		def path(name: str) -> str:
			return self.linkFlex(pyPageNode, name, True)

		env |= {'link': link}
		env |= {'path': path}

		env |= {'getLastModifiedObj': lambda: pyPageNode.lastModifiedObj}
		env |= {'getLastModified': pyPageNode.getLastModified}
		env |= {'getIdeaDateObj': pyPageNode.getIdeaDateObj}
		env |= {'getIdeaDate': pyPageNode.getIdeaDate}
		env |= {'getCreateDateObj': pyPageNode.gitFirstAuthDate}
		env |= {'getCreateDate': pyPageNode.getCreateDate}

		# Invoke pypage
		pyPageOutput = pypage(rawPyPageFileText, env)

		# Perform initial Markdown processing:
		if isinstance(pyPageNode, Md):
			PyPageNode.temporal_link = link
			mdResult = Md.processMarkdown(pyPageOutput)
			PyPageNode.temporal_link = None
			env.update(mdResult.metadata)
			pyPageOutput = mdResult.html

		# Handle `public` var:
		if 'public' in env:
			if env['public'] is True:
				pyPageNode.makePublic()
			elif env['public'] is False:
				pyPageNode.shouldPublish = False

		# Perform Markdown template application:
		if isinstance(pyPageNode, Md):
			templateHtml = self.getTemplateHtml(env)
			self.inTemplate = True
			# Re-process against `templateHtml` with PyPage:
			pyPageOutput = pypage(templateHtml, env | {'content': pyPageOutput})
			self.inTemplate = False

		# Set the PyPageNode's output:
		pyPageNode.output = pyPageOutput

		# Enrich with `env`.
		pyPageNode.env = env
		for k, v in self.getModuleVars(env).items():
			if not hasattr(pyPageNode, k):
				setattr(pyPageNode, k, v)

	def runConfigIfAny(self, dirNode: DirNode, env: dict[str, Any]) -> Dict[str, Any]:
		# Run a __config__.py file, if one exists.
		configEnv = env.copy()
		configFileL = [f for f in dirNode.files if f.fileName == Fs.configFileName]
		if configFileL:
			configFile: FileNode = configFileL[0]

			def path(name: str) -> str:
				return self.link(configFile, self.nameRegistry.lookup(name), True)

			configEnv |= {'path': path}

			print(
				f'{Fore.dark_orange}Running:{Style.reset}',
				os.path.join(dirNode.fullPath, Fs.configFileName),
			)
			exec(Fs.readfile(Fs.configFileName), configEnv)
		return configEnv

	@staticmethod
	def getSkipNames(env: dict[str, Any]) -> List[str]:
		skipNames = []
		if 'skip' in env:
			skipVar = env['skip']
			if isinstance(skipVar, list):
				for skipName in skipVar:
					if isinstance(skipName, str):
						skipNames.append(skipName)
					else:
						raise AltezaException(
							'`skip` must be a list of strings representing names to be skipped.\n'
							+ f'`{skipName}` is not a string.'
						)
			else:
				raise AltezaException('`skip` must be a list of names.')
		return skipNames

	def process(self) -> None:
		def walk(dirNode: DirNode, env: dict[str, Any]) -> None:
			env = env.copy()  # Duplicate env.
			env |= {'dir': dirNode}  # Enrich with current dir.
			env |= self.getModuleVars(self.runConfigIfAny(dirNode, env))  # Run config.
			skipNames = self.getSkipNames(env)  # Process `skip`.

			# Ordering Note: We must recurse into the subdirectories first.
			for d in dirNode.subDirs:
				if d.dirName not in skipNames:
					with enterDir(d.dirName):
						walk(d, env)

			# Ordering Note: Files in the current directory must be processed after
			# all subdirectories have been processed so that they have access to
			# information about the subdirectories.
			for pyPageNode in dirNode.getPyPagesOtherThanIndex():
				if pyPageNode.getLinkName() not in skipNames:
					self.invokePyPage(pyPageNode, env)

			# We must process the index file last.
			indexPage: Optional[PageNode] = dirNode.indexPage
			if indexPage is not None and isinstance(indexPage, PyPageNode):
				self.invokePyPage(indexPage, env)

			# TODO: Enrich dirNode with additional `env`/info from index?

		initial_env = self.seed | self.getBasicHelpers()

		walk(self.rootDir, initial_env)

		self.tracePublic()

	def tracePublic(self) -> None:
		"""Make all nodes reachable from public nodes public. (Called after processing.)"""
		if '/' in self.nameRegistry.allFiles:
			# Always make the root (/) level index page public, if it exists.
			rootIndex = self.nameRegistry.allFiles['/']
			rootIndex.makePublic()

		publicNodes: List['FsNode'] = []

		def gatherPublicNodes(fsNode: FsNode) -> None:
			if isinstance(fsNode, DirNode):
				for dirNode in itertools.chain(fsNode.subDirs, fsNode.files):
					gatherPublicNodes(dirNode)
			if fsNode.shouldPublish:
				publicNodes.append(fsNode)

		gatherPublicNodes(self.rootDir)

		print('\nInitial pre-reachability public files:')
		for node in filter(lambda pNode: isinstance(pNode, FileNode), publicNodes):
			print('/' + node.fullPath)

		seen: Set['FsNode'] = set()

		def makeReachableNodesPublic(fsNode: FsNode) -> None:
			if fsNode in seen:
				return
			seen.add(fsNode)

			if not fsNode.shouldPublish:
				fsNode.makePublic()

			for linkedToNode in fsNode.linksTo:
				makeReachableNodesPublic(linkedToNode)

		for node in publicNodes:
			makeReachableNodesPublic(node)

	@staticmethod
	def fixSysPath() -> None:
		"""
		This is necessary for import statements inside executed .py to consider the current directory.
		Without this, for example, an import statement inside a `__config__.py` file will error out.
		See: https://stackoverflow.com/questions/57870498/cannot-find-module-after-change-directory
		"""
		sys.path.insert(0, '')

	@staticmethod
	def getModuleVars(env: Dict[str, Any]) -> Dict[str, Any]:
		return {k: v for k, v in env.items() if (not k.startswith('_') and not isinstance(v, types.ModuleType))}

	@staticmethod
	def getBasicHelpers() -> Dict[str, Any]:
		return {'readfile': Fs.readfile, 'sh': sh}

	@staticmethod
	def splitPath(path: str) -> List[str]:
		head, tail = os.path.split(path)
		if head == '':
			return [path]
		return Content.splitPath(head) + [tail]

	def getTemplateHtml(self, env: dict[str, Any]) -> str:
		if 'layoutRaw' in env:
			templateRaw = env['layoutRaw']
			if not isinstance(templateRaw, str):
				raise AltezaException('The `layoutRaw` must be a string.')
			print(f'  {Fore.purple_3}Applying raw template...{Style.reset}')
			return templateRaw
		if 'layout' in env:
			templateName = env['layout']
			print(
				f'  {Fore.purple_3}Applying template: {Fore.blue_violet}{templateName}{Fore.purple_3}...{Style.reset}'
			)
			if templateName in self.templateCache:
				return self.templateCache[templateName]
			templateFile = self.nameRegistry.lookup(templateName)
			templateRaw = Fs.readfile(templateFile.absoluteFilePath)
			self.templateCache[templateName] = templateRaw
			return templateRaw
		raise AltezaException('You must define a `layout` or `layoutRaw` in some ancestral `__config__.py` file.')


@contextlib.contextmanager
def enterDir(newDir: str) -> Generator[None, None, None]:
	# https://stackoverflow.com/a/13847807/908430
	oldDir = os.getcwd()
	os.chdir(newDir)
	try:
		yield
	finally:
		os.chdir(oldDir)


class Engine:
	# Engine.generate(...) is called to write the output of a processed Content object.
	# Engine.makeSite() is called to perform a full site generation.
	# Engine.run() is used to invoke Alteza overall.
	def __init__(self, args: Args) -> None:
		self.args: Args = args
		# Just copying & renaming a few args:
		self.shouldCopyAssets: bool = args.copy_assets
		self.contentDir: str = args.content
		self.outputDir: str = args.output
		# Other instance variables:
		self.shouldExit: bool = False
		self.setIgnoreAbsPaths(args)

	@staticmethod
	def generateMdContents(md: Md) -> None:
		if os.path.exists('index.html'):
			raise AltezaException(f'An index.html already exists, and conflicts with {md}, at {os.getcwd()}.')
		with open('index.html', 'w', encoding='utf-8') as pageHtml:
			pageHtml.write(md.output)

	@staticmethod
	def generateMd(md: Md) -> None:
		if not md.isIndex():
			os.mkdir(md.realName)
			with enterDir(md.realName):
				Engine.generateMdContents(md)
		else:
			Engine.generateMdContents(md)

	@staticmethod
	def generateNonMd(nonMd: NonMd) -> None:
		fileName = nonMd.rectifiedFileName
		if os.path.exists(fileName):
			raise AltezaException(f'File {fileName} already exists, and conflicts with {nonMd}.')
		with open(fileName, 'w', encoding='utf-8') as nonMdPageFile:
			nonMdPageFile.write(nonMd.output)

	@staticmethod
	def generatePyPageNode(pyPageNode: PyPageNode) -> None:
		if isinstance(pyPageNode, Md):
			Engine.generateMd(pyPageNode)

		elif isinstance(pyPageNode, NonMd):
			Engine.generateNonMd(pyPageNode)

		else:
			raise AltezaException(f'{pyPageNode} pyPage attribute is invalid.')

	def generateStaticAsset(self, fileNode: FileNode) -> None:
		if self.shouldCopyAssets:
			shutil.copyfile(fileNode.absoluteFilePath, fileNode.fileName)
		else:
			os.symlink(fileNode.absoluteFilePath, fileNode.fileName)

	def generate(self, content: Content) -> None:
		def walk(curDir: DirNode) -> None:
			for subDir in filter(lambda node: node.shouldPublish, curDir.subDirs):
				os.mkdir(subDir.dirName)
				with enterDir(subDir.dirName):
					walk(subDir)

			for fileNode in filter(lambda node: node.shouldPublish, curDir.files):
				if isinstance(fileNode, PyPageNode):
					Engine.generatePyPageNode(fileNode)
				else:
					self.generateStaticAsset(fileNode)

		with enterDir(self.outputDir):
			walk(content.rootDir)

	def checkContentDir(self) -> None:
		if not os.path.isdir(self.contentDir):
			raise AltezaException(f"The provided path '{self.contentDir}' does not exist or is not a directory.")

	def resetOutputDir(self) -> None:
		if os.path.isfile(self.outputDir):
			raise AltezaException(
				f'A file named {self.outputDir} already exists. Please move it or delete it. '
				'Note that if this had been a directory, we would have erased it.'
			)
		if os.path.isdir(self.outputDir):
			if not self.args.clear_output_dir:
				raise AltezaException(
					f'Specified output directory {self.outputDir} already exists.\n'
					'Please use --clear_output_dir to delete it prior to site generation.'
				)
			print(f'Deleting directory {Fore.dark_red_2}%s{Style.reset} and all of its content...\n' % self.outputDir)
			shutil.rmtree(self.outputDir)
		os.mkdir(self.outputDir)

	def processContent(self) -> Content:
		with enterDir(self.contentDir):
			print('Analyzing content directory...')
			fsCrawlResult = Fs.crawl()
			print(fsCrawlResult.nameRegistry)
			content = Content(self.args, fsCrawlResult)
			print('Processing...\n')
			content.process()
			print('\nSuccessfully completed processing.\n')

		print('File Tree:')
		print(fsCrawlResult.rootDir.displayDir())

		return content

	def makeSite(self) -> None:
		startTimeNs = time.time_ns()

		self.checkContentDir()
		self.resetOutputDir()

		content = self.processContent()
		print('Generating...')
		self.generate(content)

		elapsedMilliseconds = (time.time_ns() - startTimeNs) / 10**6
		print(
			# pylint: disable=consider-using-f-string
			'\nSite generation complete (Alteza %s). Time elapsed: %.2f ms' % (alteza_version, elapsedMilliseconds)
		)

	def makeSiteWithExceptionHandling(self) -> None:
		try:
			self.makeSite()
		except (AltezaException, PypageError) as e:
			logging.exception(e)
			print('\nSite build failed due to Alteza or PyPage error.')
		except Exception as e:
			logging.exception(e)
			print('\nSite build failed.')

	@staticmethod
	def setIgnoreAbsPaths(args: Args) -> None:
		Fs.ignoreAbsPaths = []
		for somePath in args.ignore:
			if os.path.exists(somePath):
				Fs.ignoreAbsPaths.append(os.path.abspath(somePath))
			else:
				raise AltezaException(f'Path to ignore `{somePath}` does not exist.')

	class WatchdogEventHandler(FileSystemEventHandler):
		def __init__(self, contentDir: str) -> None:
			self.contentDirAbsPath: str = os.path.abspath(contentDir)
			self.timeOfMostRecentEvent: Optional[int] = None

		def on_any_event(self, event: FileSystemEvent) -> None:
			for ignoreAbsPath in Fs.ignoreAbsPaths:
				if ignoreAbsPath in event.src_path or ignoreAbsPath in event.dest_path:
					return
			if '__pycache__' in event.src_path or '__pycache__' in event.dest_path:
				return
			if Fs.isHidden(os.path.basename(os.path.normpath(event.src_path))):
				return
			if isinstance(event, DirModifiedEvent) and event.src_path == self.contentDirAbsPath:
				return

			print('Detected a change in:', event.src_path or event.dest_path)

			self.timeOfMostRecentEvent = max(self.timeOfMostRecentEvent or 0, time.time_ns())

	def runWatchdog(self) -> None:
		self.makeSiteWithExceptionHandling()

		timeIntervalNs = 2 * 10**8
		timeIntervalSecs = 0.2

		def watching() -> None:
			print('\nWatching for changes... press Ctrl+C to exit.')

		eventHandler = Engine.WatchdogEventHandler(self.contentDir)
		observer = WatchdogObserver()
		observer.schedule(eventHandler, self.contentDir, recursive=True)
		observer.start()
		try:
			watching()

			def signalHandler(sig: int, frame: Optional[types.FrameType]) -> None:
				# pylint: disable=unused-argument
				print('\nExiting...')
				self.shouldExit = True

			signal.signal(signal.SIGINT, signalHandler)

			while not self.shouldExit:
				time.sleep(timeIntervalSecs)
				if eventHandler.timeOfMostRecentEvent:
					timeSinceMostRecentEvent = time.time_ns() - (eventHandler.timeOfMostRecentEvent or 0)
					if timeSinceMostRecentEvent > timeIntervalNs:
						eventHandler.timeOfMostRecentEvent = None
						print('\nRebuilding...\n')
						try:
							self.makeSiteWithExceptionHandling()
						except AltezaException as e:
							logging.exception(e)
							print('\nSite build failed.')
						watching()
		finally:
			observer.stop()
			observer.join()

	def run(self) -> None:
		if self.args.watch:
			self.runWatchdog()
		else:
			self.makeSite()



version: str = '0.9.2'

name: str = 'alteza'

# pylint: disable=consider-using-f-string
repo_url: str = 'https://github.com/arjun-menon/%s' % name
download_url: str = '%s/archive/v%s.tar.gz' % (repo_url, version)

author: str = 'Arjun G. Menon'
author_email: str = 'contact@arjungmenon.com'



#!/usr/bin/env python3
from .engine import Engine, Args


def main() -> None:
	Engine(Args().parse_args()).run()


# See: https://chriswarrick.com/blog/2014/09/15/python-apps-the-right-way-entry_points-and-scripts/

if __name__ == '__main__':
	main()



from setuptools import setup

from alteza.version import version, name, repo_url, download_url, author, author_email

setup(
	name=name,
	version=version,
	python_requires='>=3.10',
	description='Super-flexible Static Site Generator',
	# Ref: https://packaging.python.org/en/latest/discussions/install-requires-vs-requirements/
	install_requires=[
		# Standard dependencies:
		'pypage >= 2.1.0',
		'markdown >= 3.6',
		'mdx-breakless-lists >= 1.0.1',
		'mdx_truly_sane_lists >= 1.3',
		'pygments >= 2.16.1',
		'pyyaml >= 6.0.1',
		'colored >= 2.2.4',
		'sh >= 2.0.7',
		'typed-argument-parser >= 1.10.1',
		'watchdog >= 4.0.1',
	],
	long_description=open('README.md').read(),
	long_description_content_type='text/markdown',
	url=repo_url,
	download_url=download_url,
	author=author,
	author_email=author_email,
	keywords=['static site generator', 'static sites', 'ssg'],
	license='AGPL-3.0-or-later',
	# Ref:https://docs.python.org/3.11/distutils/examples.html
	packages=['alteza'],
	entry_points={
		'console_scripts': [f'{name}=alteza.__main__:main'],
	},
	classifiers=[
		'Development Status :: 4 - Beta',
		'Intended Audience :: Developers',
		'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
		'Topic :: Internet :: WWW/HTTP :: Dynamic Content :: Content Management System',
		'Topic :: Internet :: WWW/HTTP :: Site Management',
		'Topic :: Software Development :: Libraries :: Python Modules',
		'Topic :: Text Processing :: Markup :: Markdown',
		'Topic :: Text Processing :: Markup :: HTML',
		'Operating System :: OS Independent',
		'Environment :: Console',
		'License :: OSI Approved :: GNU Affero General Public License v3',
	],
)
